package server

import (
	"fmt"
	"game/common/proto/ipb"
	"game/common/proto/pb"
	"game/common/serviceName"
	"game/common/topicName"
	"game/common/utils"
	"game/server/gate/config"
	"game/server/gate/model"
	"github.com/fox/fox/etcd"
	"github.com/fox/fox/log"
	"github.com/fox/fox/service"
	"github.com/fox/fox/ws"
	"github.com/fox/fox/xrand"
	"github.com/golang/protobuf/proto"
)

var Gates []*GateService

type GateService struct {
	*service.NatsService
	etcdService *etcd.Registry[etcd.ServiceNode]
	wss         *ws.WsServer
}

func Init() {
	for i := 0; i < config.Command.ServiceNum; i++ {
		sid := config.Command.ServiceId + i
		if gate := newGateService(sid); gate != nil {
			Gates = append(Gates, gate)
		}
	}
}

func Stop() {
	for _, gate := range Gates {
		gate.NotifyStop()
	}
	for _, gate := range Gates {
		gate.WaitStop()
	}
}

func newGateService(serviceId int) *GateService {
	var err error
	s := new(GateService)
	sName := fmt.Sprintf("%v-%d", serviceName.Gate, serviceId)
	if s.NatsService, err = service.NewNatsService(serviceName.Gate, sName, s, config.Cfg.Nats.Address...); err != nil {
		log.Fatal(err.Error())
		return nil
	}

	if s.etcdService, err = etcd.NewRegistry[etcd.ServiceNode](config.Cfg.Etcd.Address, etcd.ServiceNode{}.EtcdRootKey(), "", ""); err != nil {
		log.Error(err.Error())
		s.NatsService.OnStop()
		return nil
	}
	wsAddress := fmt.Sprintf("%v:%v", config.Cfg.Special.Address, config.Cfg.Special.Port)
	s.wss = ws.NewWsServer(wsAddress, s.WsOnMessage, s.WsOnDisconnect)
	endpoint := &etcd.ServiceNode{
		TypeId:      int(pb.ServiceTypeId_STI_Gate),
		Name:        s.Name(),
		Type:        s.Type(),
		Address:     "",
		Port:        0,
		Version:     "",
		ServiceType: etcd.Unique,
	}
	if err = s.etcdService.Register(endpoint); err != nil {
		log.Error(err.Error())
		s.NatsService.OnStop()
		return nil
	}
	s.OnInit()
	return s
}

func (s *GateService) OnInit() {
	s.etcdService.WatchServices()

	if err := s.NatsService.Subscribe(service.Topic(s)); err != nil {
		log.Error(err.Error())
	}
	// if err := s.NatsService.QueueSubscribe(service.GroupTopic(s), service.GroupQueue(s)); err != nil {
	// 	log.Error(err.Error())
	// }

	s.wss.Run()
	log.Debug("onInit")
}

func (s *GateService) OnStop() {
	s.etcdService.UnregisterService()
	s.etcdService.UnregisterService()
	s.NatsService.OnStop()
	log.Debug("OnStop")
}

// 通过handler转发消息
func (s *GateService) OnMessage(msg []byte) error {
	log.Debug(s.Log("on message:%v", string(msg)))
	return nil
}

func (s *GateService) findService(serviceTypeId pb.ServiceTypeId) *etcd.ServiceNode {
	var nodes []*etcd.ServiceNode
	var newVer string
	s.etcdService.GetNodes().Range(func(_, value interface{}) bool {
		if node, ok := value.(*etcd.ServiceNode); ok && node.TypeId == int(serviceTypeId) {
			if newVer < node.Version {
				newVer = node.Version
			}
			nodes = append(nodes, node)
		}
		return true
	})
	var newNodes []*etcd.ServiceNode
	for _, node := range nodes {
		if node.Version == newVer {
			newNodes = append(newNodes, node)
		}
	}
	if len(newNodes) == 0 {
		return nil
	}
	return nodes[xrand.IntN(len(nodes))]
}

/*
查找topic,根据serviceTypeId以及玩家id查找玩家过往访问该服务的节点，优先使用原节点
*/
func (s *GateService) findTopic(userId int64, serviceTypeId pb.ServiceTypeId) string {
	if userId != 0 {
		if sName, ok := userServiceMgr.FindServiceName(userId, serviceTypeId); ok {
			return service.TopicEx(sName)
		}
	}
	if sNode := s.findService(serviceTypeId); sNode != nil {
		return service.TopicEx(sNode.Name)
	}
	return ""
}

// 运行在conn的read协程里
func (s *GateService) WsOnMessage(conn ws.IConn, data []byte) {
	msg := &pb.ClientMsg{}
	if err := proto.Unmarshal(data, msg); err != nil {
		log.Error(err.Error())
		return
	}
	if topic := s.findTopic(conn.UserId(), msg.ServiceId); topic != "" {
		iMsg := &ipb.InternalMsg{ConnId: conn.Id(), UserId: conn.UserId(), MsgId: msg.MsgId, Msg: msg.Data}
		dMsg, _ := proto.Marshal(iMsg)
		_ = s.Send(topic, dMsg)
	}
	log.Debug(s.Log("client to gate:%v", utils.Marshal(msg)))
}

// 向内部服务发送消息
func (s *GateService) SendServiceData(topic string, conn ws.IConn, msgId int32, data []byte) {
	iMsg := &ipb.InternalMsg{ConnId: conn.Id(), UserId: conn.UserId(), MsgId: msgId, Msg: data}
	dMsg, _ := proto.Marshal(iMsg)
	_ = s.Send(topic, dMsg)
}

// 向内部服务发送消息
func (s *GateService) SendServiceMsg(topic string, conn ws.IConn, msgId int32, msg proto.Message) {
	iMsg := &ipb.InternalMsg{ConnId: conn.Id(), UserId: conn.UserId(), MsgId: msgId}
	iMsg.Msg, _ = proto.Marshal(msg)
	dMsg, _ := proto.Marshal(iMsg)
	_ = s.Send(topic, dMsg)
}

// 向玩家发送消息
func (s *GateService) SendClientMsg(conn ws.IConn, msgId int32, msg proto.Message) {
	cMsg := &pb.ClientMsg{MsgId: msgId}
	cMsg.Data, _ = proto.Marshal(msg)
	dMsg, _ := proto.Marshal(cMsg)
	_ = conn.SendMsg(dMsg)
}

/*
todo:针对顶号的情况，由其它网关通知本网关向玩家发送下线消息，玩家关闭连接后，触发WsOnDisconnect，此时WsOnDisconnect鉴定时顶号后不需要做任务特殊处理
*/
func (s *GateService) WsOnDisconnect(conn ws.IConn) {
	if conn.UserId() > 0 {
		userServiceMgr.CleanUser(conn.UserId())
		ug, err := model.NewUserGate().Get(conn.UserId(), s.Name())
		if err != nil {
			log.Error(err.Error())
		} else {
			// 相同网关，则为正常下线，删除redis信息然后向内网广播下线
			// 不同网关，异地登陆顶号 todo:异地登陆顶号，由其它网关通知本网关向玩家发送下线消息，并主动关闭连接，不广播下线消息，不会触发
			if ug.GateName == s.Name() {
				model.NewUserGate().Del(conn.UserId(), s.Name())
				s.SendServiceMsg(topicName.UserOffline, conn, int32(pb.LoginMsgId_NtfUserOffline), &pb.NtfUserOfflineMsg{UserId: conn.UserId()})
			}
		}
		log.Debug(s.Log("user:%v disconnect", conn.UserId()))
	}
}
